from wagtail.core.jinja2tags import WagtailCoreExtension

import jinja2

try:
    import django_tex
except ImportError:
    django_tex = None


from wagtail.core import blocks
from wagtail.contrib.table_block.blocks import TableBlock

import re
from html.parser import HTMLParser

from django.utils.safestring import mark_safe
from django.conf import settings


class WagtailCoreExtensionLatex(WagtailCoreExtension):
    
    def _include_block(self, value, context=None):
        """
        Automatically translate richtext blocks into latex
        """

        if isinstance(value.block, blocks.RichTextBlock) or isinstance(value.block, TableBlock):
            
            if hasattr(value, 'render_as_block'):
                if context:
                    new_context = context.get_all()
                else:
                    new_context = {}
                
                return jinja2.Markup(richtext_as_tex(value.render_as_block(context=new_context)))

        return super()._include_block(value, context=context)


def latex_escape(string):
    # Map every special latex character to its escaped version
    
    string = string.replace("\\", "\\textbackslash")
    
    for symbol in "&%$#_{}":
        string = string.replace(symbol, "\\"+symbol)
        
    string = string.replace("~", "\\textasciitilde")
    string = string.replace("^", "\\textasciicircum")

    return string


class SimpleHtmlToLatexParser(HTMLParser):
    """
    Translate HTML into LateX
    
    This is a relatively basic parser, which maps common html tags to latex
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.latex = []
        self.tex_inline = ''
        # table support
        self.table_caption = ''
        self.column_format_key = '__column_format__'
        self.table_column_counter = 0
        self.table_column_counter_max = 0
    
    # Some simple HTML Tags, which have a similar representation in latex
    mapping = {
        'i': [" \\textit{", "} "],
        'b': [" \\textbf{", "} "],
        'p': ["\n", "\n"],
        'h1': ["\n\\section{", "}\n"],
        'h2': ["\n\\subsection{", "}\n"],
        'h3': ["\n\\subsubsection{", "}\n"],
        'h4': ["\n\\paragraph{", "}\n"],
        'ul': ["\n\\begin{itemize}\n", "\n\\end{itemize}\n"],
        'ol': ["\n\\begin{enumerate}\n", "\n\\end{enumerate}\n"],
        'li': ["\\item ", "\n"],
        'br': ["\n\\\\\n", ""],
        #'img':["\\includegraphics{", "}"],
        #'a': ["\\href{", "}"],
    }
    
    def handle_starttag(self, tag, attrs):
        a = dict(attrs)
        
        # simple mapping e.g. <b> --> \textbf{
        if tag in self.mapping:
            self.latex.append(self.mapping[tag][0])
        
        # the link must be in one line i.e. WITHOUT linebreaks --> Inline mode
        if tag == "a":
            self.tex_inline = "\n\\href{"+a.get("href", '')+"}{"
        
        # simply embedding <img> as includegraphics is not very useful, as LateX does not support CSS and classes
        # For this reason <img> is translated into a command which can then be customized within latex
        if tag == "img":
            self.latex.append("\\wagtailimage{"+a.get("src", '')+"}{"+a.get("class", '')+"}")
            
        # add support for tables
        if tag == "table":
            self.latex.append("\\begin{table}[H]\n\\centering\n\\begin{tabular}{ " + self.column_format_key + " }\n\\hline\n")
        elif tag == "caption":
            self.tex_inline = "\\caption{"
        elif tag == "thead":
            pass
        elif tag == "tbody":
            pass
        elif tag == "tr":
            pass
        elif tag == "th":
            if not self.table_column_counter:
                self.latex.append("\\textbf{")
            else:
                self.latex.append("& \\textbf{")

            # Update column counter
            self.table_column_counter += 1
            # Update maximum value
            if self.table_column_counter_max < self.table_column_counter:
                self.table_column_counter_max = self.table_column_counter
        elif tag == "td":
            if not self.table_column_counter:
                pass
            else:
                self.latex.append("& ")

            # Update column counter
            self.table_column_counter += 1
            # Update maximum value
            if self.table_column_counter_max < self.table_column_counter:
                self.table_column_counter_max = self.table_column_counter

    def handle_endtag(self, tag):
        
        # End inline mode, add line to latex statements
        if self.tex_inline:
            if tag == "a":
                self.latex.append(self.tex_inline+"}\n")
                # reset inline buffer
                self.tex_inline = ''
        else:
            # simple mapping e.g. </b> --> }
            if tag in self.mapping:
                self.latex.append(self.mapping[tag][1])

        # add support for tables
        if tag == "table":
            self.latex.append("\\end{tabular}\n" + self.table_caption + "\\end{table}")
            
            # build column format based on number of columns
            column_format = '|'
            for col in range(self.table_column_counter_max):
                column_format += 'c|'

            # replace column format key
            for idx, line in enumerate(self.latex):
                if self.column_format_key in line:
                    line = line.replace(self.column_format_key, column_format)
                    break
            self.latex[idx] = line
                
        elif tag == "caption":
            self.tex_inline += "}\n"
            self.table_caption = self.tex_inline
            # reset inline buffer
            self.tex_inline = ''
        elif tag == "thead":
            pass
        elif tag == "tbody":
            pass
        elif tag == "tr":
            self.latex.append(" \\\\ \\hline\n")
            # reset column counter
            self.table_column_counter = 0
        elif tag == "th":
            self.latex.append("} ")
        elif tag == "td":
            self.latex.append(" ")

    def handle_data(self, data):
        
        # preprocess the data 
        data = latex_escape(data).strip()
        
        # inline mode (no breaks)
        if self.tex_inline:
            self.tex_inline += data
        else:
            self.latex.append(data)
            
    def parse(self, html):
        # reset
        self.latex = []
        self.tex_inline = ''
        
        # process html
        self.feed(html)
        
        latex = "".join(self.latex)
        
        # remove multiple spaces
        latex = re.sub(' +', ' ', latex)
        
        # we're outputting latex so autoescape does not make any sense.
        return mark_safe(latex)


# The user may define a custom latex parser
HTML_TO_LATEX_PARSER = getattr(settings, "HTML_TO_LATEX_PARSER", SimpleHtmlToLatexParser)

def richtext_as_tex(richtext):
    # Parse a richtext as latex
    try:
        html = richtext.__html__()
    except AttributeError:
        html = str(richtext)
    
    return HTML_TO_LATEX_PARSER().parse(html)


if django_tex:
    from django_tex.filters import FILTERS

    def latex_environment(**options):
        # Setup a Jinja2 environment usable for wagtail in latex mode
        
        if not "autoescape" in options.keys():
            options["autoescape"] = None
        
        if not "extensions" in options.keys():
            options["extensions"] = []
        
        options["extensions"].append("django_tex.extensions.GraphicspathExtension")
        options["extensions"].append(WagtailCoreExtensionLatex)
        
        # add django-tex filters and richtext filter
        env = jinja2.Environment(**options)
        env.filters.update(FILTERS)
        env.filters["richtext"] = richtext_as_tex
        
        return env




from wagtail.contrib.routable_page.models import RoutablePageMixin, route


from django.conf import settings


from django.utils.translation import gettext as _

from django.utils.cache import add_never_cache_headers

from .utils import route_function


from django.utils.cache import patch_cache_control


from wagtail.core.models import Page

# set default pdf provider (weasyprint/django-tex)
try:
    from .views import WagtailWeasyView
    DEFAULT_PDF_VIEW_PROVIDER = WagtailWeasyView
    
except ImportError:
    try:
        from .views import WagtailTexView
        DEFAULT_PDF_VIEW_PROVIDER = WagtailTexView
        
    except ImportError:
        print("Warning: DEFAULT_PDF_VIEW_PROVIDER unspecified." + 
              "Make sure you have either django-weasyprint or django-tex installed, or provide a default.")
        
        DEFAULT_PDF_VIEW_PROVIDER = None


class MultipleViewPageMixin(RoutablePageMixin):
    """
    This mixin enables multiple different views on a wagtail page.
    
    The goals of this extension are similar to the @route decorator from wagtails RoutablePage.
    In contrast to the native RoutablePageMixin this mixin is build with the goal to provided
    a more flexible/extensible inheritance structure for views on wagtail pages.
    
    With this mixin, a page model (e.g. class PdfPage(Page)) can define an class attribute
    
    ROUTE_CONFIG = [
        ("html", r'^$'),     # default route
        ("pdf", r'^pdf/$'),  # /pdf/ route
    ]
    
    which adds the methods
    >   @route(r'^$')
    >   def serve_html(..)
    and 
    >   @route(r'^pdf/$')
    >   def serve_pdf(..)
    to the class and default the classes preview modes with
    >   DEFAULT_PREVIEW_MODES = [("html", "html preview"), ("pdf", "pdf preview")]
    
    To make the difference to a naive implementation with @route is that an inheriting class
    is still able to change/extend the path configuration to its needs by reimplementing
    ROUTE_CONFIG.
    
    E.g. Another page model (e.g. class CustomPdfPage(PdfPage)) can default the "pdf" view,
    which would not be easily possible by using @route("^pdf/") in PdfPage
    
    ROUTE_CONFIG = [
        ("pdf", r'^$'),    # new default route
        ("html", None),    # ignored route
    ]
    """
    
    def __init_subclass__(cls):
        """
            Process cls.ROUTE_CONFIG on class creation.
            
            Renew cls.DEFAULT_PREVIEW_MODES to match the configuration and
            implement @routed serve_*() methods
        """
        
        # Check if the inheriting class really is a true page.
        # This prevents adding unwanted routes to mixins inheriting from this class
        if issubclass(cls, Page):
            
            cls.DEFAULT_PREVIEW_MODES = []
            
            for key, value, *args in cls.ROUTE_CONFIG: # TODO rename route config
                
                if value:
                    cls.DEFAULT_PREVIEW_MODES.append((key, cls.get_preview_name(key)))
                    
                    serve_method = "serve_{}".format(key)
                    
                    # add the @route decorator to the serve methods
                    fn = getattr(cls, serve_method)
                    setattr(cls, serve_method, route_function(fn, value, *args))
    
    @classmethod
    def get_preview_name(cls, key):
        """
        Suggested name for the preview key
        
        e.g. "pdf" will become "pdf preview"
        """
        
        return _(str(key)+" preview")
    
    @property
    def preview_modes(self):
        """
        A list of (internal_name, display_name) tuples for the modes in which
        this page can be displayed for preview/moderation purposes.
        By default this is set to a list of all available views (given by ROUTE_CONFIG),
        e.g. [("html", "html preview"), ("pdf", "pdf preview")].
        The defaults are provided by DEFAULT_PREVIEW_MODES of the class
        """
        return type(self).DEFAULT_PREVIEW_MODES

    
    # kwargs not supported (by wagtail) yet
    def serve_preview(self, request, mode_name):
        """
        Handle serve_preview like wagtail (version 2.12)
        but hook in mode specific serve_*()/serve_preview_*() methods.
        e.g. call serve_preview_pdf() or serve_pdf() for mode="pdf" 
        """
        
        mode_names = [mode[0] for mode in self.preview_modes]
        
        request.is_preview = True
        
        # try to serve preview or serve regular if the given mode is available as preview
        if mode_name in mode_names:
            try:
                serve = getattr(self, "serve_preview_" + mode_name)
            except AttributeError:
                serve = getattr(self, "serve_" + mode_name)
                
            response = serve(request)
        
            patch_cache_control(response, private=True)
            
            return response
        
        return super().serve_preview(request, mode_name)


class PdfModelMixin:
    TEMPLATE_ATTRIBUTE = 'template_name'
    
    def get_context(self, *args, **kwargs):
        return {}
    
    # TODO context=... for block
    def get_template(self, request, *args, extension=None, **kwargs):
        
        try:
            return getattr(self, self.TEMPLATE_ATTRIBUTE)
        except AttributeError:
            raise AttributeError("Template attribute not found for model {}. "
                                 "Try to add {}='path/to/your/template_file' to {}".format(self._meta.model_name, self.TEMPLATE_ATTRIBUTE, self._meta.model_name)
                                 )
    

class PdfViewPageMixin(MultipleViewPageMixin):
    """
    A mixin for serving a wagtailpage as '.pdf'
    
    This works by rerouting the pages sub-url (example.com/path/to/page/<sub-url>) with
    wagtails routable pages and rendering it with a custom pdf view.
    
    For this to work you have to ensure that you've installed 'WeasyPrint'.
    Alternatively you can also use 'django-tex' (make sure that you have 'luatex' installed)
    and either implement
    >    PDF_VIEW_PROVIDER = DjangoTexProvider
    in your page model, or add
    >    DEFAULT_PDF_VIEW_PROVIDER = DjangoTexProvider
    to your projects settings.
    
    By default only the pdf view is available, i.e. you may only view this page as pdf.
    This may be changed by reimplementing ROUTE_CONFIG, e.g.
    
    ROUTE_CONFIG = [
        ("pdf", r'^pdf/$'), # pdf view
        ("html", r'^$'),    # default view
    ]
    
    will serve the page as usual and /pdf/ will serve the rendered '.pdf' document.
    
    You should avoid to override the serve() method, as this likely will break the routing.
    """
    
    # by default only the pdf view is available, i.e. you may only view this page as pdf
    ROUTE_CONFIG = [
        ("pdf", r'^$'),
        ("html", None),
    ]
    
    # you can implement Page.attachment to control the Content-Disposition attachment
    ATTACHMENT_VARIABLE = "attachment"
    
    PDF_VIEW_PROVIDER = getattr(settings, "DEFAULT_PDF_VIEW_PROVIDER", DEFAULT_PDF_VIEW_PROVIDER)
    
    
    def get_pdf_view(self, **kwargs):
        """
        Get the serve method for the classes pdf provider
        """
        
        return self.PDF_VIEW_PROVIDER.as_view()
    
    def get_pdf_filename(self, request, **kwargs):
        """
        Get the filename for the pdf file
        
        This simply extends the page title with '.pdf'
        """
        
        return self.title + '.pdf'
    
    
    def get_template(self, request, extension=None, **kwargs):
        """
        Get the template name for this page
        
        extension can be used to replace the file extension '.html' with e.g. '.tex'
        """
        
        template_name = super().get_template(request)
        
        if extension:
            template_name = template_name.replace(".html", "."+extension)
            
        return template_name
    
    
    def make_preview_request(self, original_request=None, preview_mode=None, extra_request_attrs=None):
        """
        Make a preview request with the orignial request (admin view) still being available
        
        This is essentially a fix for weasyprint in the wagtail admin preview.
        The original request is still accessible as request.orginal_request e.g. to figure out
        the server port (which is not possible otherwise, as wagtail is creating a new 'fake'
        request with port 80)
        """
        
        if not extra_request_attrs:
            extra_request_attrs = {}
            
        extra_request_attrs["original_request"] = original_request
        
        return super().make_preview_request(original_request=original_request, preview_mode=preview_mode, extra_request_attrs=extra_request_attrs)
    
    def serve_html(self, request, **kwargs):
        return super().serve(request)
    
    def serve_pdf(self, request, **kwargs):
        """
            Serve the page as pdf using the classes pdf view
        """
        
        view = self.get_pdf_view(**kwargs)
        response = view(request, object=self, mode="pdf", **kwargs)
        
        response['Content-Disposition'] = '{}filename="{}"'.format(
            "attachment;" if getattr(self, self.ATTACHMENT_VARIABLE, False) else '',
            self.get_pdf_filename(request, **kwargs)
        )
        
        # TODO remove
        add_never_cache_headers(response)
            
        return response





from functools import wraps
from wagtail.contrib.routable_page.models import route


def route_function(func, pattern, *args, **kwargs):
    """
    Adds the @route decorator to func
    """
    
    @wraps(func) 
    @route(pattern, *args, **kwargs)
    def inner(*args, **kwargs):
        return func(*args, **kwargs)
    
    return inner
    



import setuptools

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='wagtail-pdf-view',
    version='0.1.3',
    description='PDF rendering views for the Wagtail CMS',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jonas Donhauser',
    #author_email='',
    url='https://github.com/donhauser/wagtail-pdf',
    packages=['wagtail_pdf_view'],
    package_data={'': ['LICENSE', 'templates/*']},
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=["wagtail"],
    extras_require = {
        'weasyprint':["django-weasyprint"],
        'django-tex':["django-tex"],
    },
)





from django.template.response import TemplateResponse
from django.views.generic.base import ContextMixin, TemplateResponseMixin, View
from django.views.generic.detail import SingleObjectMixin, BaseDetailView

from django.contrib.staticfiles.finders import find

from django.conf import settings

from django.core.exceptions import SuspiciousFileOperation


class ConcreteSingleObjectMixin(SingleObjectMixin):
    """
    This mixin simply enables you to pass a concrete instance of a object to a DetailView
    instead of fetching it with a Queryset.
    
    In the regular django dataflow this a DetailView needs to match the correct object
    e.g. given its primary key. When the concrete object is already given (e.g. during the
    serve process of a wagtail page) fetching the same object from the database again is
    non-optimal. With this mixin the object may simply passed to the view call with
    'object=..'
    """
    
    def setup(self, request, *args, object=None, **kwargs):
        self.object = object
        
        return super().setup(request, *args, **kwargs)

    def get_object(self):
        return self.object or super().get_object()

class WagtailAdapterMixin(ContextMixin):
    """
    Make a wagtail page accessible as template view.
    
    A wagtail page already provides a template name and context for rendering.
    This mixin makes the page rendering information available for the use with a plain
    django view (TemplateResponse).
    """
    
    def get_template_names(self):
        if hasattr(self.object, "get_template"):
            return self.object.get_template(self.request, view_provider=self)
        
        # fallback
        return super().get_template_names()
    
    def get_context_data(self, **kwargs):
        context = self.object.get_context(self.request, view_provider=self, **kwargs)
        context.update(super().get_context_data(**kwargs))
        return context
    

try:
    import django_tex
except ImportError:
    django_tex = None

if django_tex:
    
    from django_tex.core import compile_template_to_pdf

    class TexTemplateResponse(TemplateResponse):
        
        @property
        def rendered_content(self):
            """
            Returns rendered PDF pages.
            """
            
            context = self.resolve_context(self.context_data)
            
            return compile_template_to_pdf(self.template_name, context)


    class WagtailTexTemplateMixin(WagtailAdapterMixin, ConcreteSingleObjectMixin, TemplateResponseMixin):
        """
        Provide the latex compiler (from django-tex) as view
        """
        content_type='application/pdf'
        response_class = TexTemplateResponse
        
        
        def get_template_names(self):
            if hasattr(self.object, "get_template"):
                return self.object.get_template(self.request, view_provider=self, extension="tex")
            
            # fallback
            return super().get_template_names()


    class WagtailTexView(WagtailTexTemplateMixin, BaseDetailView):
        pass


try:
    import django_weasyprint
except ImportError:
    django_weasyprint = None


if django_weasyprint:

    import weasyprint
    from django_weasyprint.views import WeasyTemplateResponseMixin, WeasyTemplateResponse

    class CustomWeasyTemplateResponse(WeasyTemplateResponse):
        
        def get_base_url(self):
            """
            Determine base URL to fetch CSS files from `WEASYPRINT_BASEURL` or
            fall back to using the root path of the URL used in the request.
            
            In contrast to the implementation in django_weasyprint this method contains a
            working fallback for the dummy requests used by wagtails page preview mode.
            """
            
            if hasattr(settings, 'WEASYPRINT_BASEURL'):
                return settings.WEASYPRINT_BASEURL
            
            # Check if this is a wagtail dummy request and use the uri of the original request instead
            if getattr(self._request, "is_dummy", False) and hasattr(self._request, "original_request"):
                return self._request.original_request.build_absolute_uri('/')
                
            return self._request.build_absolute_uri('/')


        def get_css(self, base_url, url_fetcher):
            """
            Get the css for weasyprint
            
            All paths are collected from _stylesheets and are tried to be located
            with djangos static loaders.
            """
            
            tmp = []
            for value in self._stylesheets:
                try:
                    # Try to import relative to BASE_DIR
                    css = weasyprint.CSS(
                        value,
                        base_url=base_url,
                        url_fetcher=url_fetcher,
                    )
                except FileNotFoundError as e:
                    try:
                        # Try to locate the static file
                        path = find(value)
                    except SuspiciousFileOperation:
                        # raise original FileNotFound, raising SuspiciousFileOperation would be misleading
                        raise e
                    
                    if path:
                        css = weasyprint.CSS(
                            path,
                            base_url=base_url,
                            url_fetcher=url_fetcher,
                        )
                    else:
                        raise e
                    
                if css:
                    tmp.append(css)
                
            return tmp


    class WagtailWeasyTemplateMixin(WagtailAdapterMixin, ConcreteSingleObjectMixin, WeasyTemplateResponseMixin):
        response_class = CustomWeasyTemplateResponse
        
        def get_pdf_stylesheets(self):
            # try to call get_stylesheets, otherwise get stylesheet attribute
            try:
                stylesheets = self.object.get_stylesheets(self.request)
            except AttributeError:
                stylesheets = getattr(self.object, "stylesheets", [])
            
            return stylesheets


    class WagtailWeasyView(WagtailWeasyTemplateMixin, BaseDetailView):
        def get(self, request, *args, **kwargs):
            self.object = self.get_object()
            
            kwargs["object"] = self.object
            
            # Difference to BaseDetailView: Also pass kwargs
            context = self.get_context_data(**kwargs)
            
            return self.render_to_response(context)
        
        # support for post (e.g. for filling forms)
        def post(self, request, *args, **kwargs):
            return self.get(self, request, *args, **kwargs)

 
